# -*- coding: utf-8 -*-

from werkzeug.urls import url_encode

from odoo import api, fields, models, _
from odoo.addons.openerp_enterprise.models.spoof import no_impostor
from odoo.addons.saas_worker.const import WEBSITE
from odoo.exceptions import UserError

class Project(models.Model):

    _inherit = 'project.project'

    enterprise_need = fields.Boolean('Need Enterprise Subscription', help="Tasks/Issues of this project need to have a valid Enterprise subscription to be processed")
    show_task_description = fields.Boolean(
        'Show Task Description (Portal)',
        help="If checked, the task description will be visible on the portal",
        default=True,
    )


class Task(models.Model):

    _inherit = 'project.task'

    reviewer_id = fields.Many2one('res.users', 'Reviewer', index=True, default=lambda self: self.env.user, tracking=True)
    enterprise_need = fields.Boolean(related='project_id.enterprise_need')
    can_request_staff_access = fields.Boolean(compute="_compute_can_request_staff_access")

    # TODO rename without mnt_
    # It is now used for more than maintenance. As soon it is set, we use it as the DB of the ticket
    # for listing DB, connect to support page, ...
    mnt_subscription_id = fields.Many2one('sale.order', string="Subscription",
                                          help="Subscription set on /help form",
                                          domain=[('is_subscription', '=', True)])

    @api.model
    def _message_get_auto_subscribe_fields(self, updated_fields, auto_follow_fields=None):
        if auto_follow_fields is None:
            auto_follow_fields = ['user_id', 'reviewer_id']
        return super(Task, self)._message_get_auto_subscribe_fields(updated_fields, auto_follow_fields)

    @api.depends('enterprise_need', 'enterprise_subscription_ids', 'mnt_subscription_id')
    def _compute_can_request_staff_access(self):
        granted_spoof_subscriptions = self.env['openerp.enterprise.spoof.authorization'].sudo().search([
            ('user_id', '=', self.env.user.id),
            ('state', '=', 'accepted'),
            ('expire_date', '>=', fields.Datetime.now()),
        ]).mapped('subscription_id')

        self.can_request_staff_access = False
        for task in self:
            if task.enterprise_need:
                subscription = task._get_valid_db_subscription_on_ticket()
                if subscription:
                    task.can_request_staff_access = (
                        subscription.is_staff_restricted
                        and subscription not in granted_spoof_subscriptions
                    )
    def action_connect_support_page(self):
        self.ensure_one()
        if self.saas_database_url:
            main_database = self.enterprise_db_ids[:1]
            return main_database.with_context(task_id=self.id).action_connect_support_page()
        return True  # don't do anything

    def _compute_saas_database_url(self):
        for ticket in self:
            ticket.saas_database_url = ''
            if ticket.project_id.enterprise_need:
                if len(ticket.enterprise_db_ids) == 1 and ticket.enterprise_db_ids.hosting in ("saas", "paas"):
                    ticket.saas_database_url = ticket.enterprise_db_ids.url

    def _compute_enterprise_db_ids(self):
        for issue in self:
            contracts = self.env['sale.order']
            databases = self.env['openerp.enterprise.database']
            if issue.project_id.enterprise_need:
                if issue.mnt_subscription_id:
                    contracts = issue.mnt_subscription_id
                else:
                    if issue.enterprise_subscription_ids:
                        contracts = issue.enterprise_subscription_ids
                    elif issue.partner_id and issue.partner_id.commercial_partner_id:
                        contracts = self.env['sale.order'].search([
                            ('partner_id', '=', issue.partner_id.commercial_partner_id.id),
                            ('is_subscription', '=', True)
                        ])
                databases = contracts.mapped('database_ids')
            issue.enterprise_db_ids = databases

    enterprise_subscription_ids = fields.One2many(related='partner_id.enterprise_subscription_ids', string="Enterprise Subscriptions")
    enterprise_db_ids = fields.One2many(
        'openerp.enterprise.database', string='Enterprise Databases', compute='_compute_enterprise_db_ids'
    )
    enterprise_open_issue_ids = fields.One2many(
        'project.task', string="Enterprise Open Issues", compute='_compute_enterprise_open_issues'
    )
    subscription_state = fields.Selection(related='mnt_subscription_id.subscription_state', string="Subscription State")
    subscription_to_renew = fields.Boolean(related='mnt_subscription_id.to_renew')
    saas_database_url = fields.Char(compute='_compute_saas_database_url')

    @api.depends('partner_id', 'project_id')
    def _compute_enterprise_open_issues(self):
        projects = (self.env.ref('openerp_enterprise.project_help')
                    | self.env.ref('openerp_enterprise.project_customer_upgrade'))
        for record in self:
            open_issues = self.env['project.task']
            if record.project_id in projects:
                commercial_partner = record.partner_id.commercial_partner_id
                contract_partners = record.enterprise_subscription_ids.partner_id

                open_issues = self.search([
                    ('project_id', '=', record.project_id.id),
                    ('partner_id', 'child_of', (commercial_partner | contract_partners).ids),
                    ('stage_id.fold', '=', False),
                ]) - record

            record.enterprise_open_issue_ids = open_issues

    def action_show_customer_issues(self):
        self.ensure_one()
        if self.project_id not in (
                self.env.ref('openerp_enterprise.project_help')
                | self.env.ref('openerp_enterprise.project_customer_upgrade')):
            return False

        commercial_partner = self.partner_id.commercial_partner_id or self.partner_id
        contract_partners = self.enterprise_subscription_ids.partner_id

        return {
            'type': 'ir.actions.act_window',
            'name': _("Customer's Issues"),
            'res_model': 'project.task',
            'view_mode': 'tree,form',
            'domain': [
                ('project_id', '=', self.project_id.id),
                ('partner_id', 'child_of', (commercial_partner | contract_partners).ids),
            ],
            'context': self._context,
        }

    def _message_auto_subscribe_notify(self, partner_ids, *args, **kwargs):
        user = self.env.user
        if user.partner_id.id in partner_ids:
            partner_ids.remove(user.partner_id.id)
        super(Task, self)._message_auto_subscribe_notify(partner_ids, *args, **kwargs)

    def _message_post_after_hook(self, *args, **kwargs):
        """ After Odoo 10.0 we have "merged" project.issue and project.task.
            By doing so we lost some "features", one of them was that after each message posted
            on the issue, we were updating the field "date_action_last" with the current date.
            Which has the side effect of updating the write_date of the issue.

            For the project Help, we have an automated action that automatically set to the "done" stage
            the tickets that are in the "Customer Feedback" for more than 7 days (by using the write_date).
            Which means, if a customer replied by mail to the ticket, the write_date was updated,
            thus leaving the ticket for a week more on the "customer feedback" stage.

            Since we lost this feature, we have to manually update the write date ourselves,
        """
        self.env.cr.execute("UPDATE project_task SET write_date=now() AT TIME ZONE 'UTC' WHERE id = any(%s)", (self.ids,))
        self.invalidate_recordset(['write_date'])
        return super(Task, self)._message_post_after_hook(*args, **kwargs)

    def _get_valid_db_subscription_on_ticket(self):

        self.ensure_one()
        partner_sub_tpl = self.env.ref('openerp_enterprise.quotation_template_odoo_partnership')
        if self.mnt_subscription_id and self.mnt_subscription_id.sale_order_template_id != partner_sub_tpl:
            return self.mnt_subscription_id
        if len(self.enterprise_subscription_ids) == 1:
            sub = self.enterprise_subscription_ids
            return sub if sub.sale_order_template_id.id != partner_sub_tpl.id and len(sub.database_ids) == 1 else None
        return None

    @no_impostor
    def ask_for_subscription_access(self):
        self.ensure_one()
        if not self.can_request_staff_access:
            raise UserError('You cannot request access to a subscription with no staff restrictions')

        subscription = self._get_valid_db_subscription_on_ticket()  # As partner_subscription are not subject to staff restriction, we should only have one subscription
        if not subscription:
            raise UserError('You cannot request an access if a task is not associated to a valid subscription.')

        composer_form_view_id = self.env.ref('mail.email_compose_message_wizard_form').id
        template_id = self.env.ref('openerp_enterprise.subscription_task_request_access')
        # no translation, that's fine, it's a tech audit log
        reason = f"ticket: [#{self.id}] {self.display_name}"
        auth_vals = {
            'subscription_id': subscription.id,
            'reason': reason,
            'user_id': self.env.user.id,
            'state': 'draft',
            'responsible_partner_id': subscription.partner_id.id,
        }
        spoof_auth = self.env['openerp.enterprise.spoof.authorization'].sudo().create(auth_vals)
        req_url = WEBSITE + '/my/subscription/access-form?' + url_encode({
            'id': spoof_auth.id,
            'token': spoof_auth._get_authorization_token(),
            'task_id': self.id,
        })
        ctx = {
            'default_composition_mode': 'comment',
            'default_res_ids': [self.id],
            'default_model': 'project.task',
            'default_template_id': template_id.id,
            'default_partner_ids': [subscription.partner_id.id],
            'active_ids': self.ids,
            'subscription_name': subscription.display_name,
            'sub_access_request': True,
            'request_url': req_url,
            'spoof_id': spoof_auth.id,
        }
        return {
            'type': 'ir.actions.act_window',
            'view_mode': 'form',
            'res_model': 'mail.compose.message',
            'view_id': composer_form_view_id,
            'target': 'new',
            'context': ctx,
        }

    @api.returns('mail.message', lambda value: value.id)
    def message_post(self, **kwargs):
        if self.env.context.get('sub_access_request'):
            subscription = self._get_valid_db_subscription_on_ticket()
            database = subscription.database_ids[0]  # Normally we should only have one database on the subscription
            # no translation, that's fine, it's a tech audit log
            reason = f"ticket: [#{self.id}] {self.display_name}"
            spoof_vals = {
                'user_id': self.env.user.id,
                'db_uuid': database.db_uuid,
                'db_id': database.id,
                'action': "access_request",
                'reason': reason,
                'success': True,
                'restricted': False,  # No restrictions to ask database access
            }
            self.env['openerp.enterprise.spoof'].sudo().create(spoof_vals)
            # Body is already computed so, changing the request url in the context does not update it.
            self.env['openerp.enterprise.spoof.authorization'].sudo().browse(self.env.context.get('spoof_id')).write({'state': 'pending'})
            self._message_log(body="Access requested to subscription: %s" % (subscription.display_name))
        return super(Task, self).message_post(**kwargs)

    def _notify_by_email_add_values(self, base_mail_values):
        if self.env.context.get('sub_access_request'):
            base_mail_values.update(is_notification=False)
        return super(Task, self)._notify_by_email_add_values(base_mail_values)
